#!/usr/local/bin/ruby

# rwiki.rb
#
# $Id: rwiki.rb,v 1.82 2002/10/15 15:19:32 seki Exp $
# Copyright (c) 2000-2001 Masatoshi SEKI
#
# rwiki.rb is copyrighted free software by Masatoshi SEKI.
# You can redistribute it and/or modify it under the same term as Ruby.

$KCODE = 'EUC'

require 'rw-config'

require 'mutexm'
require 'erb/erbu'
require 'erb/erbl'
require 'rd/rdfmt'
require 'rd2rwiki-lib'

require 'rw-lib'
require 'hotpage'

require 'drb/drb'

require 'md5'

module RWiki

  Version.regist('rwiki server', '1.2.4 ($Revision: 1.82 $) ext')

  class PageStore
    def []=(*arg)
      k = arg.shift
      v = arg.pop
      rev = arg.shift
      check_revision(k, rev)
      set(k, store(v))
    end

    def [](k)
      retrieve(get(k))
    end

    def modifed(k)
      nil
    end

    def revision(k)
      modified(k).to_s
    end

    def check_revision(k, rev)
      return unless rev
      raise RevisionError, [rev, revision(k)].inspect unless rev == revision(k)
    end

    private
    def store(s)
      s
    end

    def retrieve(s)
      s
    end
  end

  class PageStoreFile < PageStore
    include MutexM
    def initialize(path)
      @dir = path.split(File::Separator)
      super()
    end

    private
    def set(k,v,rev=nil)
      synchronize do
	return if v.nil?
	begin
	  fp = File.open(fname(k), 'w')
	  fp.write(v)
	ensure
	  fp.close if fp
	end
      end
    end

    def get(k)
      synchronize do
	begin
	  fp = File.open(fname(k), 'r')
	  fp.read
	rescue Errno::ENOENT
	  return nil
	ensure
	fp.close if fp
	end
      end
    end

    public
    def modified(k)
      begin
	File.stat(fname(k)).mtime
      rescue Errno::ENOENT
	nil
      end
    end

    def revision(k)
      make_digest(get(k))
    end

    private
    def escape(str)
      str.gsub(/([^a-zA-Z0-9_-])/n){ sprintf("%%%02X", $1.unpack("C")[0]) }
    end

    def fname(k)
      File.join(*(@dir + [escape(k)]))
    end

    def make_digest(src)
      MD5.new(src).hexdigest
    end
  end

  class Content
    def initialize(src)
      @body = nil
      @links = []
      set_src(src)
    end
    attr_reader(:body, :links, :src)
    
    private
    def set_src(src)
      @src = src
      @body = src
    end
  end

  class ContentRD2 < Content
    include ERbUtil

    private
    def set_src(src)
      if src
	begin
	  @src = src
	  tree = RD::RDTree.new("=begin\n#{@src}\n=end\n")
	  v = RD::RD2RWikiVisitor.new
	  @body = v.visit(tree)
	  @links = v.links
	rescue
	  @body = "<h1>RD Error</h1>\n"
	  @body << "<pre>#{h($!)}</pre>\n"
	  @body << "<pre>\n"
	  cnt = 2	# '=begin' is the first line.
	  @src.each_line do |line|
	    @body << "%4d| %s" % [ cnt, h(line) ]
	    cnt += 1
	  end
	  @body << "</pre>\n"
	end
      end
    end
  end

  class PageFormat
    include ERbUtil
    extend ERbMethod

    @@address = ADDRESS
    @@mailto = MAILTO
    @@css = CSS
    @@title = TITLE
    @@lang = LANG || KCode.lang
    @@charset = CHARSET || KCode.charset

    def PageFormat.navi_view(name, pg, format)
      %Q[<a href="#{ format.ref_name(name) }">#{ format.h name }</a>]
    end

    def modified(t)
      return '-' unless t
      dif = (Time.now - t).to_i
      dif = dif / 60
      return "#{dif}m" if dif <= 60
      dif = dif / 60
      return "#{dif}h" if dif <= 24
      dif = dif / 24
      return "#{dif}d"
    end

    def initialize(env = {}, &block)
      @env = env
      @block = block
    end

    def var(key)
      @block ? @block.call(key) : nil
    end

    def env(key)
      @env[key]
    end

    def ref_url(url)
      h(url)
    end

    def ref_name(name, cmd = 'view')
      page_url =
	if env('ref_name')
	  env('ref_name').call(cmd, name)
	else
	  program = env('base')
	  req = Request.new(cmd, name)
	  program + "?" + req.query
	end
      ref_url(page_url)
    end

    def full_ref_name(name, cmd = 'view')
      page_url = "#{env('base_url')}?#{Request.new(cmd, name).query}"
      ref_url(page_url)
    end

    def form_action
      env('base')
    end

    def form_hidden(name, cmd = 'view')
      # req = Request.new(cmd, name) ???
      %Q[<input type="hidden" name="cmd" value="#{cmd}" />] +
        %Q[<input type="hidden" name="name" value="#{h(name)}" />]
    end

    def_erb_method('header(pg)', 'header.rhtml')
    def_erb_method('navi(pg)', 'navi.rhtml')

    def body(pg, opt = {})
      # ERbLight.new(pg.body.to_s, 4).result(binding)
      str = ERbLight.new(pg.body.to_s).result(binding)
      if opt.has_key?(:key)
	# Copy keys from UI side.
	keys = opt[:key].collect { |i| i.dup }
	str = hilighten(str, keys)
      end
      str
    end

    def_erb_method('footer(pg)', 'footer.rhtml')

    def_erb_method('view(pg)', 'view.rhtml')
    def_erb_method('edit(pg)', 'edit.rhtml')
    def_erb_method('submit(pg)', 'submit.rhtml')
    def_erb_method('emphasize(pg)', 'emphasize.rhtml')
    def_erb_method('error(pg)', 'err.rhtml')
    def_erb_method('src(pg)', 'src.rhtml')

    def_erb_method('map_view(pg)', 'map.rhtml')
    def_erb_method('list_view(pg)', 'list.rhtml')
    def_erb_method('recent_view(pg)', 'recent.rhtml')
    def_erb_method('search_view(pg)', 'search.rhtml')

    private
    # erbu does not define 'u' method...
    def escape_url(s)
      s.to_s.gsub(/[^a-zA-Z0-9_\-.]/n){ sprintf("%%%02X", $&.unpack("C")[0]) }
    end
    alias u escape_url

    def hilighten(str, keywords)
      hilighted = str.dup
      keywords.each do |key|
	re = Regexp.new('(' << Regexp.escape(h(key)) << ')', Regexp::IGNORECASE)
	hilighted.gsub!(/([^<]*)(<[^>]*>)?/) {
	  body, tag = $1, $2
	  body.gsub(re) {
	    %Q[<em style="color:black;background-color:#ffff66">#{$1}</em>]
	  } << ( tag || "" )
	}
      end
      hilighted
    end
  end

  class Page
    include DRbUndumped
    extend Enumerable

    @@format = PageFormat
    @@db = PageStoreFile.new(DB_DIR)
    @@content = ContentRD2

    @@dirty = false
    @@extra_page = []

    @@fw_table = {}
    FW_MUTEX = Object.new
    FW_MUTEX.extend(MutexM)
    FW_MUTEX.taint

    def Page.[](name)
      FW_MUTEX.synchronize do 
	obj = @@fw_table[name]
	return obj if obj
	obj = new(name)
	@@fw_table[name] = obj
	obj.update_src(@@db[name])
	obj
      end
    end

    def Page.active_names(ary, root)
      return if ary[root] 
      return unless @@fw_table[root]
      ary[root] = @@fw_table[root]
      ary[root].links.each do |name|
	Page.active_names(ary, name) unless ary[name]
      end
    end

    def Page.gc
      FW_MUTEX.synchronize do
	return unless @@dirty
	@@dirty = false
	ary = {}
	@@extra_page.each do |root|
	  Page.active_names(ary, root)
	end
	@@fw_table = ary
      end
    end

    def Page.each
      FW_MUTEX.synchronize do
	@@fw_table.each_value do |page|
	  yield page
	end
      end
    end

    def Page.size
      @@fw_table.size
    end

    def Page.find_title(str)
      self.find_all { |page| page.name.downcase.index(str.downcase) }
    end

    # Bare in mind that the argument regexp must be trusted.
    def Page.search_title(regexp)
      self.find_all { |page| regexp =~ page.name }
    end

    def Page.find_body(str)
      self.find_all { |page| page.contain? str }
    end

    # Bare in mind that the argument regexp must be trusted.
    def Page.search_body(regexp)
      self.find_all { |page| page.match? regexp }
    end

    class << self
      alias find_str find_body
    end

    def Page.protect_gc(name)
      @@extra_page.push name unless @@extra_page.include? name
    end

    def Page.include_name?(name)
      @@fw_table.include?(name)
    end

    private_class_method :new

    def initialize(name)
      @name = name
      @src = nil
      @body = nil
      @links = nil
      @revlinks = []
      @modified = @@db.modified(name)
      @format = @@format

      @hot_order = self.method(:hot_order)
      @hot_links = HotPage.new( &@hot_order )
      @hot_revlinks = HotPage.new( &@hot_order )
      @hot_edges = HotPageContainer.new( &@hot_order )
      @hot_edges << @hot_links << @hot_revlinks
    end
    attr_reader(:name, :src, :body, :links, :revlinks, :modified)
    attr_accessor(:format)

    def revision
      @@db.revision(@name).to_s
    end

    def src=(v)
      set_src(v, nil)
    end

    def set_src(v, rev)
      FW_MUTEX.synchronize do
	@@dirty = true
	@@db[@name, rev] = v
	@modified = Time.now
	update_src(v)
      end
      Page.gc
    end

    def empty?
      src.nil? || src == ''
    end

    def contain?(str)
      return false if empty?
      src.downcase.index(str.downcase) ? true : false
    end

    def match?(regexp)
      return false if empty?
      regexp =~ src
    end

    def view_html(env = {}, &block)
      @format.new(env, &block).view(self)
    end

    def edit_html(env = {}, &block)
      @format.new(env, &block).edit(self)
    end

    def submit_html(env = {}, &block)
      @format.new(env, &block).submit(self)
    end

    def emphatic_html(env = {}, &block)
      @format.new(env, &block).emphasize(self)
    end

    def error_html(env = {}, &block)
      @format.new(env, &block).error(self)
    end
    
    def src_html(env = {}, &block)
      @format.new(env, &block).src(self)
    end

    def body_html(env = {}, &block)
      @format.new(env, &block).body(self)
    end

    def hot_links
      @hot_links.pages.uniq
    end

    def hot_revlinks
      @hot_revlinks.pages.uniq
    end

    def hot_edges
      @hot_edges.pages.uniq
    end

    # For backward compatibility.
    def hotlinks
      hot_edges.collect { |name| RWiki::Page[name] }
    end

    def update_src(v)
      obsolete_links

      content = @@content.new(v)
      @src = content.src
      @body = content.body
      @links = content.links
      @hot_links.replace(@links)

      update_links
    end

    protected
    def add_link(from)
      @hot_links.set_dirty
    end

    def del_link(from)
    end

    def add_rev_link(from)
      unless @revlinks.include? from
	@revlinks.push from
	@hot_revlinks << from
      end
    end

    def del_rev_link(from)
      @revlinks.delete(from)
      @hot_revlinks.delete(from)
    end

    private
    def update_links
      return unless @links
      @links.each do |to|
	Page[to].add_rev_link(@name) unless to == @name
      end
      @revlinks.each do |from|
	Page[from].add_link(@name) unless from == @name
      end
    end

    def obsolete_links
      return unless @links
      @links.each do |to|
	Page[to].del_rev_link(@name) unless to == @name
      end
      @revlinks.each do |from|
	Page[from].del_link(@name) unless from == @name
      end
    end

    def <=>(rhs)
      @name <=> rhs.name
    end

    Paste = Time.at(0)
    def hot_order( a, b )
      a = RWiki::Page[a].modified || Paste
      b = RWiki::Page[b].modified || Paste
      b <=> a
    end
  end

  class Front
    include DRbUndumped

    def include?(name)
      Page.include_name?(name)
    end

    def find(str)
      Page.find_body(str).collect do |pg|
        pg.name
      end
    end

    def find_all(str)
      Page.find_body(str).each do |pg|
        yield pg
      end
    end

    def view(name, env = {}, &block)
      page = Page[name]
      page.empty? ? page.edit_html(env, &block) : page.view_html(env, &block)
    end

    def edit_view(name, env = {}, &block)
      Page[name].edit_html(env, &block)
    end

    def submit_view(name, env = {}, &block)
      Page[name].submit_html(env, &block)
    end

    def emphatic_view(name, env = {}, &block)
      Page[name].emphatic_html(env, &block)
    end

    def error_view(name, env = {}, &block)
      Page[name].error_html(env, &block)
    end

    def src(name)
      page = Page[name]
      page.empty? ? nil : page.src
    end

    def body(name)
      page = Page[name]
      page.empty? ? nil : page.body
    end
    
    def modified(name)
      Page[name].modified
    end
    
    def src_view(name, env = {}, &block)
      Page[name].src_html(env, &block)
    end

    def set_src_and_view(name, src, env = {}, &block)
      page = Page[name]
      page.src = src
      page.view_html(env, &block)
    end

    def page(name)
      Page[name]
    end
  end

  NAVI = []
  def install_page_module(name, format, navi = false)
    pg = Page[name]
    pg.format = format
    Page.protect_gc(name)
    NAVI.push([name, format]) if navi
  end
  module_function :install_page_module
end

if __FILE__ == $0

  unless $DEBUG
    # Run as a daemon...
    exit!( 0 ) if fork
    Process.setsid
    exit!( 0 ) if fork
  end

  top = RWiki::Page[RWiki::TOP_NAME] 	# build page tree
  help = RWiki::Page['help']	

  search = RWiki::Page['search']
  def search.view_html(env = {}, &block)
    @format.new(env, &block).search_view(self)
  end

  RWiki::Page.protect_gc(RWiki::TOP_NAME)
  RWiki::Page.protect_gc('help')
  RWiki::Page.protect_gc('search')

  require 'rw-info'
  require 'rw-map'
  require 'rw-orphan'
  require 'rw-like'
#  require 'rw-concat'
#  require 'rw-arb'

  DRb.start_service(RWiki::DRB_URI, RWiki::Front.new)

  if $DEBUG
    gets
    exit
  else
    STDIN.reopen('/dev/null')
    STDOUT.reopen('/dev/null', 'w')
    STDERR.reopen('/dev/null', 'w')
    DRb.thread.join
  end
end
